import logging

from concurrent_tasks import PeriodicTask

from local_tuya.errors import CommandTimeoutError
from local_tuya.events import EventNotifier
from local_tuya.protocol.events import (
    CommandSent,
    ConnectionBroken,
    ConnectionEstablished,
    ConnectionLost,
)
from local_tuya.protocol.message import HeartbeatCommand

logger = logging.getLogger(__name__)


class Heartbeat(PeriodicTask):
    def __init__(self, interval: float, event_notifier: EventNotifier):
        super().__init__(interval, _heartbeat, event_notifier)
        event_notifier.register(ConnectionLost, lambda _: self.cancel())
        event_notifier.register(ConnectionEstablished, lambda _: self.create())

    def __enter__(self) -> "Heartbeat":
        """Don't start automatically, only when connection is established."""
        return self


async def _heartbeat(event_notifier: EventNotifier) -> None:
    try:
        await event_notifier.emit(CommandSent(HeartbeatCommand()))
    except CommandTimeoutError:
        logger.warning("timeout waiting for heartbeat response")
        # This would typically happen when connection succeeded,
        # but it is impossible to communicate.
        await event_notifier.emit(ConnectionBroken())



from typing import Dict, Optional, Set, Tuple

from local_tuya.device.enums import DataPoint
from local_tuya.protocol import Value, Values

Blacklist = Dict[str, Optional[Set[Value]]]
_Blacklist = Dict[str, Set[Value]]


class Constraint:
    """Represents other values that cannot be applied if the datapoint/value is set."""

    def __init__(
        self,
        data_point: DataPoint,
        value: Value,
        *blacklist: Tuple[DataPoint, Optional[Set[Value]]],
    ):
        self._data_point = data_point
        self._value = value
        self._blacklist: Blacklist = {dp: v for dp, v in blacklist}

    def blacklist(self, values: Values) -> Blacklist:
        if values[self._data_point] != self._value:
            return {}
        return self._blacklist


class Constraints:
    """Represent all constraints for a given device."""

    def __init__(self, *constraints: Constraint):
        self._constraints = constraints

    def _blacklist(self, values: Values) -> _Blacklist:
        blacklist: _Blacklist = {}
        for constraint in self._constraints:
            for k, v in constraint.blacklist(values).items():
                if k not in blacklist:
                    blacklist[k] = set()
                if v:
                    blacklist[k] |= v
        return blacklist

    def filter_values(self, values: Values, current: Values) -> Values:
        """Filter values that can be updated given the device constraints."""
        # Check on merged values.
        blacklist = self._blacklist({**current, **values})
        filtered: Values = {}
        for data_point, value in values.items():
            if data_point in blacklist and (
                not blacklist[data_point] or value in blacklist[data_point]
            ):
                continue
            filtered[data_point] = value
        return filtered
